@@ -18,6 +18,7 @@ import os
 import time
 import warnings
 from collections.abc import Iterable, Mapping
+from copy import deepcopy
 
 import arviz as az
 import jax
@@ -27,7 +28,6 @@ import numpyro
 import pandas as pd
 import psutil
 from deprecated import deprecated
-
 from jax import config
 from mpire import WorkerPool
 from numpyro.distributions import Distribution
@@ -35,33 +35,29 @@ from numpyro.infer import MCMC, NUTS
 from scipy.optimize import curve_fit
 from tqdm.auto import tqdm
 
-from autoeis import io, julia_helpers, metrics, parser, utils
-from autoeis.models import circuit_regression_complex
+from autoeis import io, julia_helpers, metrics, models, parser, utils
+from autoeis.julia_helpers import ec, jl
 
 from .utils import InferenceResult
 
-# Enforce FP64, otherwise circuit fitter fails (because FP32 gradients not sufficient?)
-config.update("jax_enable_x64", True)
-# Tell JAX to use CPUs to avoid the annoying "GPU might be present" warning
-config.update("jax_platforms", "cpu")
-# EIS datasets are not big-enough -> CPU much faster than GPU
+# FP32 gradients not sufficient for ECM fitting
+numpyro.enable_x64()
+# GPU is actually slower for typical EIS dataset sizes
 numpyro.set_platform("cpu")
+# Get rid of the JAX warning about GPU might be available
+config.update("jax_platforms", "cpu")
 
 warnings.filterwarnings("ignore", category=Warning, module="arviz.*")
 log = logging.getLogger(__name__)
 
-# Initialize Julia runtime
-os.environ["PYTHON_JULIACALL_AUTOLOAD_IPYTHON_EXTENSION"] = "no"
-julia_helpers.ensure_julia_deps_ready(quiet=True)
-jl = julia_helpers.init_julia(quiet=True)
-ec = julia_helpers.import_backend(jl)
-
 __all__ = [
     "perform_full_analysis",
     "generate_equivalent_circuits",
     "filter_implausible_circuits",
     "perform_bayesian_inference",
     "compute_fitness_metrics",
+    "jl",
+    "ec",
 ]
 
 
@@ -125,6 +121,7 @@ def generate_equivalent_circuits(
     parallel: bool = True,
     generations: int = 30,
     population_size: int = 100,
+    terminals: str = "RLP",
     seed: int = None,
 ) -> pd.DataFrame:
     """Generates candidate circuits that fit the impedance data using
@@ -141,13 +138,16 @@ def generate_equivalent_circuits(
     complexity : int, optional
         Complexity of the ECM search space (default is 12).
     tol : float, optional
-        Convergence threshold for the ECM search (default is 1e-2).
+        Convergence threshold for the ECM search (default is 1e-32).
     parallel : bool, optional
         If True, the ECM search will be performed in parallel (default is True).
     generations : int, optional
         Number of generations for the ECM search (default is 30).
     population_size : int, optional
         Number of ECMs to generate per generation (default is 100).
+    terminals : str, optional
+        Circuit components to consider (default is "RLP"). R: resistor,
+        L: inductor, C: capacitor, P: constant-phase element.
     seed : int, optional
         Random seed for reproducibility (default is None).
 
@@ -163,14 +163,14 @@ def generate_equivalent_circuits(
 
     ec_kwargs = {
         "head": complexity,
-        "terminals": "RLP",
+        "terminals": terminals,
         "convergence_threshold": tol,
         "generations": generations,
         "population_size": population_size,
     }
 
     ecm_generator = _generate_ecm_parallel_julia if parallel else _generate_ecm_serial
-    circuits = ecm_generator(Z, freq, iters, ec_kwargs, seed)
+    circuits = ecm_generator(freq, Z, iters, ec_kwargs, seed)
 
     # Convert output to DataFrame with columns ("circuitstring", "Parameters")
     circuits = io.parse_ec_output(circuits)
@@ -193,7 +193,7 @@ def _generate_ecm_serial(
     jl.seval(f"import Random; Random.seed!({seed})")
 
     circuits = []
-    for _ in tqdm(range(iters), desc="Circuit Evolution"):
+    for _ in tqdm(range(iters), desc="Generating Candidate ECMs", leave=False):
         utils.flush_streams()
         try:
             circuit = ec.circuit_evolution(Z, freq, **ec_kwargs, quiet=True)
@@ -235,7 +235,7 @@ def _generate_ecm_parallel_julia(
     # Perform parallelized GEP in chunks
     circuits = []
 
-    with tqdm(total=iters, desc="Circuit Evolution", miniters=1) as pbar:
+    with tqdm(total=iters, desc="Generating Candidate ECMs", miniters=1, leave=False) as pbar:
         utils.flush_streams()
         for iters_ in iters_chunked:
             try:
@@ -305,50 +305,6 @@ def _generate_ecm_parallel_mpire(
     return circuits
 
 
-def _generate_ecm_parallel_mpire2(
-    freq: np.ndarray[float],
-    Z: np.ndarray[complex],
-    iters: int,
-    ec_kwargs: dict,
-    seed: int,
-):
-    """Generates candidate circuits that fit the impedance data, in parallel
-    via Python multiprocessing.
-    """
-
-    def circuit_evolution(freq, Z, ec_kwargs, seed):
-        """Closure to generate a single circuit to be used with multiprocessing."""
-        jl = julia_helpers.init_julia()
-        ec = julia_helpers.import_backend(jl)
-        # Set random seed for reproducibility
-        jl.seval(f"import Random; Random.seed!({seed})")
-        circuit = ec.circuit_evolution(Z, freq, **ec_kwargs)
-        # circuit = seed
-        return circuit
-
-    mpire_kwargs = {
-        "progress_bar": True,
-        "desc": "Circuit Evolution",
-        "timeout": 1,
-    }
-
-    # Set a different seed for each process
-    seed = [seed + i for i in range(iters)]
-
-    circuits = utils.distribute_task(
-        circuit_evolution,
-        freq,
-        Z,
-        ec_kwargs,
-        seed,
-        static=(0, 1, 2),  # static args = (freq, Z, ec_kwargs)
-        **mpire_kwargs,
-    )
-
-    # circuits = [str(c) for c in circuits if c is not None]
-    return circuits
-
-
 def split_components(circuits: pd.DataFrame) -> pd.DataFrame:
     """Augments the circuits dataframe with additional columns for each component."""
     # Initialize lists to populate the component columns later
@@ -452,37 +408,45 @@ def compute_fitness_metrics(
     circuits["converged"] = [result.converged for result in results]
     circuits["divergences"] = [result.num_divergences for result in results]
 
-    # Compute WAIC and add to the dataframe
-    waic_re = [az.waic(x, var_name="obs_real", scale="deviance") for x in mcmcs]
-    waic_im = [az.waic(x, var_name="obs_imag", scale="deviance") for x in mcmcs]
-    circuits["WAIC (real)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_re]
-    circuits["WAIC (imag)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_im]
-
     # Compute the posterior predictive and add to the dataframe
-    # NOTE: axis=1 because posterior is of shape (num_samples, num_obs)
-    fn = lambda r: utils.eval_posterior_predictive(r.InferenceResult.samples, r.circuitstring, freq)  # fmt: off
-    circuits["Z_pred"] = circuits.apply(fn, axis=1)
+    Z_pred = [utils.eval_posterior_predictive(r.samples, r.circuit, freq) for r in results]
+    circuits["Z_pred"] = Z_pred
+
+    # FIXME: This function doesn't work with Bode as objective function
+    # TODO: Extract log-likelihood of mag/phase, and turn it into that for real/imag
+
+    # Compute WAIC and add to the dataframe
+    if "sigma.real" in mcmcs[0].get_samples().keys():
+        waic_re = [az.waic(x, var_name="obs.real", scale="deviance") for x in mcmcs]
+        waic_im = [az.waic(x, var_name="obs.imag", scale="deviance") for x in mcmcs]
+        circuits["WAIC (real)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_re]
+        circuits["WAIC (imag)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_im]
+    else:
+        waic_mag = [az.waic(x, var_name="obs.mag", scale="deviance") for x in mcmcs]
+        waic_phase = [az.waic(x, var_name="obs.phase", scale="deviance") for x in mcmcs]
+        circuits["WAIC (mag)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_mag]
+        circuits["WAIC (phase)"] = [x["elpd_waic"] + 2 * x["p_waic"] for x in waic_phase]
 
     # Compute R^2 and add to the dataframe
-    fn = lambda r: metrics.r2_score(Z.real, r.Z_pred.real, axis=1)
+    fn = lambda row: metrics.r2_score(Z.real, row.Z_pred.real, axis=1)
     circuits["R^2 (real)"] = circuits.apply(fn, axis=1)
-    fn = lambda r: metrics.r2_score(Z.imag, r.Z_pred.imag, axis=1)
+    fn = lambda row: metrics.r2_score(Z.imag, row.Z_pred.imag, axis=1)
     circuits["R^2 (imag)"] = circuits.apply(fn, axis=1)
     # Since R^2 is a vector (num_samples), also compute its mean for convenience
     circuits["R^2 (ravg)"] = circuits["R^2 (real)"].apply(np.mean)
     circuits["R^2 (iavg)"] = circuits["R^2 (imag)"].apply(np.mean)
 
     # Compute MAPE and add to the dataframe
-    fn = lambda r: metrics.mape_score(Z.real, r.Z_pred.real, axis=1)
+    fn = lambda row: metrics.mape_score(Z.real, row.Z_pred.real, axis=1)
     circuits["MAPE (real)"] = circuits.apply(fn, axis=1)
-    fn = lambda r: metrics.mape_score(Z.imag, r.Z_pred.imag, axis=1)
+    fn = lambda row: metrics.mape_score(Z.imag, row.Z_pred.imag, axis=1)
     circuits["MAPE (imag)"] = circuits.apply(fn, axis=1)
     # Since MAPE is a vector (num_samples), also compute its mean for convenience
     circuits["MAPE (ravg)"] = circuits["MAPE (real)"].apply(np.mean)
     circuits["MAPE (iavg)"] = circuits["MAPE (imag)"].apply(np.mean)
 
     # Add number of parameters to the dataframe
-    circuits["n_params"] = circuits.apply(lambda r: len(r.Parameters), axis=1)
+    circuits["n_params"] = circuits.apply(lambda row: len(row.Parameters), axis=1)
 
     return circuits
 
@@ -523,36 +487,43 @@ def _validate_impedance_data(freq, Z):
     return freq, Z
 
 
-def _validate_p0(p0):
-    """Validates the initial guess for the circuit parameters."""
-    if p0 is None:
-        return None
-    # Single initial guess
-    if isinstance(p0, (Mapping, Iterable)):
-        # return [p0]
-        return p0
-    # Multiple initial guesses
-    if isinstance(p0, list):
-        for p0_ in p0:
-            assert isinstance(p0_, (Mapping, Iterable)), "Invalid initial guess."
-        return p0
-    raise ValueError("`p0` must be a dict, list[dict], or list[np.ndarray].")
+def _validate_p0(p0, broadcast_to: int = 1):
+    """Validates the initial guess for ECM parameters, and broadcasts to a list."""
+    # Validate input data type
+    if not (isinstance(p0, (dict, list)) or utils.is_ndfarray_like(p0) or p0 is None):
+        raise ValueError("'p0' must be an ndarray/dict or list[ndarray/dict].")
+
+    # If single p0 is provided, broadcast
+    if p0 is None or utils.is_ndfarray_like(p0) or isinstance(p0, dict):
+        p0 = [deepcopy(p0) for _ in range(broadcast_to)]
+
+    # If a list of p0 is provided, validate
+    assert isinstance(p0, list), "Something went wrong. Please report!"
+    for p0_ in p0:
+        msg = f"Invalid p0: {p0_}"
+        assert utils.is_ndfarray_like(p0_) or isinstance(p0_, dict), msg
+
+    return p0
 
 
-def _validate_priors(priors):
-    """Validates the priors for the circuit parameters."""
-    if priors is None:
-        return None
-    msg = "'priors' must be a dict of NumPyro distributions."
-    # Single prior
-    if isinstance(priors, Mapping):
-        assert all(isinstance(v, Distribution) for v in priors.values()), msg
-        return priors
-    # Multiple priors
-    if isinstance(priors, Iterable):
-        assert all(isinstance(prior, Mapping) for prior in priors), msg
-        return priors
-    raise ValueError("'priors' must be a dict[Distribution] or list[dict[Distribution]].")
+def _validate_priors(priors, broadcast_to: int = 1):
+    """Validates the priors for the ECM parameters, and broadcasts to a list."""
+    # Validate input data type
+    if not isinstance(priors, (Mapping, list, type(None))):
+        raise ValueError("'priors' must be a dict or list[dict].")
+
+    # If single priors is provided, broadcast
+    if priors is None or isinstance(priors, Mapping):
+        priors = [deepcopy(priors) for _ in range(broadcast_to)]
+
+    # Ensure all priors are valid
+    assert isinstance(priors, list), "Something went wrong. Please report!"
+    for priors_ in priors:
+        msg = "'priors' must be a dict of NumPyro Distribution objects."
+        assert isinstance(priors_, Mapping), msg
+        assert all(isinstance(v, Distribution) for v in priors_.values()), msg
+
+    return priors
 
 
 def _validate_seed(seed, num_splits=1) -> list[jax.Array] | jax.Array:
@@ -573,6 +544,34 @@ def _validate_seed(seed, num_splits=1) -> list[jax.Array] | jax.Array:
     return subkey
 
 
+def _refine_p0(p0, circuit, datasets, progress_bar):
+    """"""
+    if p0 is None:
+        p0 = [None] * len(datasets)
+    if len(p0) != len(circuit) != len(datasets):
+        raise ValueError("Length of 'p0', 'circuit', and 'datasets' must be the same.")
+
+    max_iters = 10
+    p0 = utils.distribute_task(
+        utils.fit_circuit_parameters,
+        circuit,
+        [dataset.freq for dataset in datasets],
+        [dataset.Z for dataset in datasets],
+        p0,
+        max_iters,
+        static=(4),  # static args = (circuit, freq, Z, max_iters)
+        progress_bar=progress_bar,
+        desc="Refining Initial Guess",
+    )
+
+    # Handle failed refinements
+    for i, elem in enumerate(p0):
+        if isinstance(elem, Exception):
+            raise RuntimeError(f"Failed to refine p0 for circuit {circuit[i]}: {elem}")
+
+    return p0
+
+
 def perform_bayesian_inference(
     circuit: str | Iterable[str] | pd.DataFrame,
     freq: np.ndarray[float] | Iterable[float],
@@ -586,10 +585,11 @@ def perform_bayesian_inference(
     num_samples: int = 1000,
     num_chains: int = 1,
     seed: int | jax.Array = None,
+    method: str = "bode",
     progress_bar: bool = True,
-    refine_p0: bool = False,
+    refine_p0: bool = True,
     parallel: bool = True,
-) -> list[InferenceResult] | InferenceResult:
+) -> list[InferenceResult]:
     """Performs Bayesian inference on the circuits based on impedance data.
 
     Parameters
@@ -620,26 +620,31 @@ def perform_bayesian_inference(
         Number of MCMC chains (default is 1).
     seed : int, optional
         Random seed for reproducibility (default is None).
+    method : str, optional
+        Objective function to minimize (default is "bode"). Choose from "bode",
+        "nyquist", "magnitude", or "chi-squared".
     progress_bar : bool, optional
         If True, a progress bar will be displayed (default is True).
     refine_p0 : bool, optional
         If True, the initial guess for the circuit parameters will be refined
-        using the circuit fitter (default is False).
+        using the circuit fitter (default is True).
     parallel : bool, optional
         If True, the MCMC chains will be run in parallel (default is True).
 
     Returns
     -------
-    InferenceResult | list[InferenceResult]
-        InferenceResult object. If multiple circuits are provided (with a
-        single freq/Z pair), or if multiple freq/Z pairs are provided (with a
-        single circuit), a list of InferenceResult objects will be returned.
+    list[InferenceResult]
+        List of InferenceResult object(s). Regardless of whether a single or
+        multiple circuits/datasets are provided, the output will always be a
+        list of InferenceResult objects for consistency.
 
     Notes
     -----
-    You cannot provide multiple circuits and multiple datasets together, i.e.,
-    either pass a single circuit with multiple datasets or multiple circuits
-    with a single dataset.
+    Only the following combinations of inputs are supported:
+    - Single circuit, single dataset
+    - Single circuit, multiple datasets
+    - Multiple circuits, single dataset
+
     """
     # Validate inputs data types
     circuit = _validate_circuit(circuit)
@@ -663,40 +668,13 @@ def perform_bayesian_inference(
     circuit = circuit * num_inferences if len(circuit) == 1 else circuit
     datasets = datasets * num_inferences if len(datasets) == 1 else datasets
 
-    # Deal with p0 or priors
-    if (p0 is not None) and (priors is not None):
-        log.warning("Both 'p0' and 'priors' are provided. Ignoring 'p0'.")
-    if priors is not None:
-        priors = _validate_priors(priors)
-        priors = [priors] * num_inferences if isinstance(priors, Mapping) else priors
-    else:
-        # Compute p0 if needed
-        if p0 is not None:
-            p0 = _validate_p0(p0)
-            p0 = [p0] * num_inferences if isinstance(p0, Mapping) else p0
-        if p0 is None or refine_p0:
-            iters = 25
-            freq, Z = zip(*[(dataset.freq, dataset.Z) for dataset in datasets])
-            args = circuit, freq, Z, [p0] * num_inferences, iters
-            p0 = utils.distribute_task(
-                utils.fit_circuit_parameters,
-                *args,
-                static=(4),  # static args = (iters)
-                progress_bar=progress_bar,
-                desc="Refining p0",
-            )
-            for i, elem in enumerate(p0):
-                if isinstance(elem, Exception):
-                    log.error(f"Failed to refine p0 for circuit {circuit[i]}: {elem}")
-                    p0[i] = None
-
-        # Convert p0 to dict if Iterable is provided
-        for i, (p0_, circuit_) in enumerate(zip(p0, circuit)):
-            if isinstance(p0_, list):
-                p0[i] = {k: v for k, v in zip(parser.get_parameter_labels(circuit_), p0_)}
-
-        # Compute priors from p0; if p0 is None, initialize priors to None
-        priors = [None if p0_ is None else utils.initialize_priors(p0_) for p0_ in p0]
+    # Ensure p0 and priors are valid and ready for parallel inference
+    p0 = _validate_p0(p0, broadcast_to=num_inferences)
+    if refine_p0 or (p0[0] is None):
+        p0 = _refine_p0(p0, circuit, datasets, progress_bar)
+    if priors is None:
+        priors = [utils.initialize_priors(p0_) for p0_ in p0]
+    priors = _validate_priors(priors, broadcast_to=num_inferences)
 
     # Generate N random seeds (one for each inference)
     seed = _validate_seed(seed, num_splits=num_inferences)
@@ -704,7 +682,7 @@ def perform_bayesian_inference(
     seed = [seed] if not isinstance(seed, list) else seed
 
     # Perform Bayesian inference
-    if parallel:
+    if parallel and num_inferences > 1:
         results = _perform_bayesian_inference_MCMD(
             circuit=circuit,
             dataset=datasets,
@@ -712,12 +690,13 @@ def perform_bayesian_inference(
             num_warmup=num_warmup,
             num_samples=num_samples,
             num_chains=num_chains,
+            method=method,
             seed=seed,
             progress_bar=progress_bar,
         )
     else:
         results = []
-        for i in range(num_inferences):
+        for i in tqdm(range(num_inferences), desc="Running Bayesian Inference", leave=False):
             result = _perform_bayesian_inference_SCSD(
                 circuit=circuit[i],
                 freq=datasets[i].freq,
@@ -727,12 +706,12 @@ def perform_bayesian_inference(
                 num_samples=num_samples,
                 num_chains=num_chains,
                 seed=seed[i],
-                progress_bar=progress_bar,
+                method=method,
+                progress_bar=False,
             )
             results.append(result)
 
-    # Return a single InferenceResult object if n_inferences = 1, else a list
-    return results[0] if len(results) == 1 else results
+    return results
 
 
 def _perform_bayesian_inference_SCSD(
@@ -740,48 +719,21 @@ def _perform_bayesian_inference_SCSD(
     freq: np.ndarray[float],
     Z: np.ndarray[complex],
     priors: Mapping[str, Distribution],
-    num_warmup: int = 2500,
-    num_samples: int = 1000,
-    num_chains: int = 1,
-    seed: int | jax.Array = None,
+    num_warmup: int,
+    num_samples: int,
+    num_chains: int,
+    method: str,
+    seed: int | jax.Array,
     progress_bar: bool = False,
 ) -> InferenceResult:
-    """Performs Bayesian inference on the circuit based on impedance data.
-
-    Parameters
-    ----------
-    circuit : str
-        CDC string representation of the input circuit. See
-        `here <https://autodial.github.io/AutoEIS/circuit.html>`_ for details.
-    freq : np.ndarray[float]
-        Frequencies corresponding to the impedance data.
-    Z : np.ndarray[complex]
-        Complex impedance data.
-    priors: Mapping[str, Distribution], optional
-        Priors for the circuit parameters (default is None).
-    num_warmup : int, optional
-        Number of warmup samples for the MCMC (default is 2500).
-    num_samples : int, optional
-        Number of samples for the MCMC (default is 1000).
-    num_chains : int, optional
-        Number of MCMC chains (default is 1).
-    seed : int | jax.Array, optional
-        Random seed for reproducibility (default is None).
-    progress_bar : bool, optional
-        If True, a progress bar will be displayed (default is False).
-
-    Returns
-    -------
-    InferenceResult
-        InferenceResult object.
-    """
     log.info("Performing Bayesian inference on the circuit {circuit}.")
 
     subkey = _validate_seed(seed)
     circuit_fn = utils.generate_circuit_fn(circuit, jit=True)
+    method = method.replace("-", "_")
 
     nuts_kernel = NUTS(
-        model=circuit_regression_complex,
+        model=getattr(models, f"circuit_regression_{method}"),
         init_strategy=numpyro.infer.init_to_median,
     )
     kwargs_mcmc = {
@@ -822,6 +774,7 @@ def _perform_bayesian_inference_MCMD(
     num_warmup: int,
     num_samples: int,
     num_chains: int,
+    method: str,
     seed: jax.Array,
     progress_bar: bool,
 ) -> list[InferenceResult]:
@@ -843,6 +796,7 @@ def _perform_bayesian_inference_MCMD(
             num_warmup=num_warmup,
             num_samples=num_samples,
             num_chains=num_chains,
+            method=method,
             seed=seed[0],
             progress_bar=False,
         )
@@ -858,9 +812,10 @@ def _perform_bayesian_inference_MCMD(
             num_warmup,  # 4
             num_samples,  # 5
             num_chains,  # 6
-            seed,  # 7
+            method,  # 7
+            seed,  # 8
             # --- distribute_task kwargs ---
-            static=(4, 5, 6),  # indices of static args
+            static=(4, 5, 6, 7),  # indices of static args
             progress_bar=progress_bar,
             desc="Performing Bayesian Inference",
         )
